/*************************************
 * treadahead - Tim's readahead
 *
 * Simple program to readahead pages on a Linux system
 *
 * Author: Tim Bird
 * Copyright 2010 Sony Network Entertainment, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * NOTES:
 * This program can readahead whole files specified on the command line, or
 * file segments (and whole files) specified in a "readlist" file.
 *
 * If the -f parameter is specified, the list of file segments to readahead
 * is read from the file /readahead_list.txt
 *
 * Each line in the readahead_list file specifies a file to read, with
 * syntax:
 * <filename>[ area_list]
 *
 * area_list is: <offset>:<len>[,<offset>:<len>]
 *
 * If the area_list is empty, then the whole file is read ahead.
 *
 * Here is an example:
 * /system/framework/framework-res.apk
 * /system/app/Contacts.apk 0:4096,1282048:4096,1294336:4096,1470464:12288,1507328:61440
 *
 * This would read everything in framework-res.apk, and only certain portions of Contacts.apk
 *
 * CHANGELOG:
 *
 * ToDo:
 ************************************/
#define _GNU_SOURCE
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/syscall.h>

#define VERSION_MAJOR	1
#define VERSION_MINOR	3	

#define READLIST_PATH	"/readahead_list.txt"
#define MAX_PATH	512
#define MAX_AREA_LIST	2048

#define STATS_OUTFILE1	"/cache/debug/d1.txt"
#define STATS_OUTFILE2	"/cache/debug/d2.txt"

extern ssize_t readahead(int fd, long long offset, size_t count); 

int verbose = 0;

void usage()
{
	printf("treadahead version %d.%d\n\n", VERSION_MAJOR, VERSION_MINOR);
	printf("Usage: treadahead [options] [<file1> [<file2> ...]]\n");
	printf("  -f   Read list of files from %s\n", READLIST_PATH);
	printf("  -f<file>   Read list of files from <file>\n");
	printf("  -h   Show this usage help\n");
	printf("  -v   Show verbose messages\n");
	printf("  -w   Re-write blocks to be preloaded, in the filesystem\n");
	printf("       This may optimize the blocks for preloading\n");
	printf("  -i   Reduce IO scheduler priority for treadahead to 'idle' (ionice -c 3)\n");
	printf("  -j   Reduce IO scheduler priority for treadahead to 'best effort- low'\n");
	printf("  -n   Set noop scheduler for sda, prior to doing readahead\n");
	printf("  -p   Reduce scheduler priority for treadahead (nice 19)\n");
	printf("  -l   Output log message on start and end\n");
	printf("  -e   Do early exit after forking, so calling program can continue.\n");
	printf("       This will redirect stdout to /dev/null, so with -v most messages are lost.\n");
	printf("  -V   Show version message\n");
	printf("  -s   Save /proc/diskstats at start and end of operation.\n");
	printf("       Files are saved to %s and %s.\n", STATS_OUTFILE1, STATS_OUTFILE2);
	exit(1);
}

#define IOPRIO_WHO_PROCESS	1
#define IOPRIO_CLASS_IDLE	3
#define IOPRIO_CLASS_BE		2	
#define IOPRIO_CLASS_SHIFT	13

void ioprio_set_idle()
{
	pid_t pid;
	int ioprio;
	int ccode;

	/* I think that the 'idle' priority class ignores the */
	/* actual priority, but set it to 7 just in case */
	ioprio = 7 | IOPRIO_CLASS_IDLE << IOPRIO_CLASS_SHIFT;

	pid = getpid();
	ccode = syscall(SYS_ioprio_set, IOPRIO_WHO_PROCESS, pid, ioprio);
	if (ccode<0) {
		printf("ccode of %d doing ioprio_set on pid %d\n", ccode, pid);
	} else {
		if (verbose) {
			printf("Successfully set ioprio to idle for pid %d\n", pid);
		}
	}
}

void ioprio_set_be_low()
{
	pid_t pid;
	int ioprio;
	int ccode;

	ioprio = 7 | IOPRIO_CLASS_BE << IOPRIO_CLASS_SHIFT;

	pid = getpid();
	ccode = syscall(SYS_ioprio_set, IOPRIO_WHO_PROCESS, pid, ioprio);
	if (ccode<0) {
		printf("ccode of %d doing ioprio_set on pid %d\n", ccode, pid);
	} else {
		if (verbose) {
			printf("Successfully set ioprio to idle for pid %d\n", pid);
		}
	}
}


void set_scheduler(char *new_scheduler)
{
	int fd;

	fd = open("/sys/block/sda/queue/scheduler", O_RDWR);
	if( fd<0 ) {
		printf("problem opening scheduler sys file\n");
	}
	write(fd, new_scheduler, strlen(new_scheduler)+1);
	close(fd);
}

void write_log(char *message)
{
	int fd;

	fd = open("/dev/kmsg", O_RDWR);
	if( fd<0 ) {
		printf("problem opening /dev/kmsg\n");
		return;
	}
	write(fd, message, strlen(message)+1);
	close(fd);
}

#define MAX_BUFFER 8192
void save_stats(char *outfile)
{
	int in_fd, out_fd;
	char buffer[MAX_BUFFER];
	int count;
	
	in_fd = open("/proc/diskstats", O_RDONLY);
	if (in_fd<0) {
		printf("problem opening /proc/diskstats\n");
		return;
	}
	out_fd = open(outfile, O_WRONLY|O_CREAT|O_TRUNC, S_IRWXU|S_IRWXG|S_IROTH);
	if (out_fd<0) {
		printf("problem opening %s\n", outfile);
		perror(NULL);
		goto err_out1;
	}
	count = read(in_fd, buffer, MAX_BUFFER);
	if (count <=0 ) {
		perror("reading stats");
		goto err_out2;
	}
	count = write(out_fd, buffer, count);
	
err_out2:
	close(out_fd);
err_out1:
	close(in_fd);
}

unsigned long total=0;

void do_readahead(char *filename, char *area_list)
{
	int fd;
	struct stat buf;
	int ccode;
	int arg_count;
	long long start_arg;
	off64_t start;	/* long long on ARM */
	//off_t start;	/* long long on ARM */
	//long long start;
	size_t len;	/* unsigned int on ARM */
	char *str1, *token, *saveptr1;

	fd = open(filename, O_RDONLY);
	if (fd<0) {
		printf("Couldn't open %s for readahead\n", filename);
		return;
	}
	fstat(fd, &buf);
	if (!area_list ) {
		ccode = readahead(fd, 0, buf.st_size);
		if (verbose)
		    printf("ccode of %d doing readahead on file %s\n", ccode, filename);
		total += buf.st_size;
		return;
	}
	/* area_list is a sequence of start:len[,start,len]... */
	/* e.g. 4096:8192,16384:4096 */
	for (str1 = area_list; ; str1 = NULL) {
		token = strtok_r(str1, ",", &saveptr1);
		if (token == NULL)
		    break;
		arg_count = sscanf(token, "%lld:%u", (long long *)&start_arg, (unsigned *)&len);
		start = ((start_arg&0xffffffff00000000LL) >>32) |
			((start_arg&0x00000000ffffffffLL) <<32) ;
		ccode = readahead(fd, start, len);
		if (verbose) {
	            printf("ccode of %d doing readahead at %lld for len %u on file %s\n",
			ccode, start, len, filename);
		}
		total += len;
	}
}

#define RW_BUF_SIZE	4096*20
char *rw_buf[RW_BUF_SIZE];

int readwrite(int fd, off_t start, size_t len)
{
	off_t cur;
	size_t remainder, this;
	ssize_t count;

	cur = start;
	remainder = len;
	while (cur < start+len) {
		this = remainder;
		if (this>RW_BUF_SIZE) {
			this = RW_BUF_SIZE;
		}
		lseek(fd, cur, SEEK_SET);
		count = read(fd, rw_buf, this);
		if(count<this) {
			printf("Could not read data at %lld for len %u - aborting\n", cur, this);
			return -1;
		}
		lseek(fd, cur, SEEK_SET);
		count = write(fd, rw_buf, this);
		if(count<this) {
			printf("Had problems writing data at %lld for len %u (only wrote %u bytes)\n", cur, this, count);
			return -1;
		}
		remainder -= this;
		cur += this;
	}
	return 0;
}

void do_readwrite(char *filename, char *area_list)
{
	int fd;
	struct stat buf;
	int ccode;
	int arg_count;
	//off64_t start;	/* long long on ARM */
	//off_t start;	/* long long on ARM */
	long long start_arg;
	off_t start;
	size_t len;	/* unsigned int on ARM */
	char *str1, *token, *saveptr1;

	fd = open(filename, O_RDWR);
	if (fd<0) {
		printf("Couldn't open %s for readwrite\n", filename);
		return;
	}
	fstat(fd, &buf);
	if (!area_list ) {
		ccode = readwrite(fd, 0, buf.st_size);
		if (verbose)
		    printf("ccode of %d doing readwrite on file %s\n", ccode, filename);
		total += buf.st_size;
		return;
	}
	/* area_list is a sequence of start:len[,start,len]... */
	/* e.g. 4096:8192,16384:4096 */
	for (str1 = area_list; ; str1 = NULL) {
		token = strtok_r(str1, ",", &saveptr1);
		if (token == NULL)
		    break;
		arg_count = sscanf(token, "%lld:%u", (long long *)&start_arg, (unsigned *)&len);
		start = (off_t)start_arg;
		if (start+len > buf.st_size) {
			len = buf.st_size-start;
			printf("hit length of file, adjusting len to %u\n", len);
		}
		ccode = readwrite(fd, start, len);
		if (verbose)
	            printf("ccode of %d doing readwrite at %lld for len %u on file %s\n",
			ccode, start_arg, len, filename);
		total += len;
	}
}

char filename_buf[MAX_PATH];
char area_buf[MAX_AREA_LIST];

int main(int argc, char **argv)
{
	int fd, arg_count;
	int i;
	int pid;
	char *filename;
	FILE *readlist;
	int log_flag=0;
	int rewrite_blocks=0;
	int new_nice;
	int stats_flag=0;
	int ioidle_flag=0;
	int iolow_flag=0;

	if (argc<2 ) {
		usage();
	}
	for( i=1; i<argc; i++ ) {
	    if (argv[i][0]=='-' && argv[i][1]=='h') {
		usage();
	    }
	    if (argv[i][0]=='-' && argv[i][1]=='v') {
		verbose = 1;
		continue;
	    }
	    if (argv[i][0]=='-' && argv[i][1]=='n') {
		set_scheduler("noop");
		printf("scheduler set to 'noop'\n");
		continue;
	    }
	    if (argv[i][0]=='-' && argv[i][1]=='i') {
		ioprio_set_idle();
		ioidle_flag=1;
		continue;
	    }
	    if (argv[i][0]=='-' && argv[i][1]=='j') {
		ioprio_set_be_low();
		iolow_flag=1;
		continue;
	    }
	    if (argv[i][0]=='-' && argv[i][1]=='p') {
		new_nice = nice(19);
		printf("new nice=%d\n", new_nice);
		continue;
	    }
	    if (argv[i][0]=='-' && argv[i][1]=='w') {
		rewrite_blocks = 1;
		continue;
	    }
	    if (argv[i][0]=='-' && argv[i][1]=='l') {
		log_flag = 1;
		write_log("treadahead start\n");
		continue;
	    }
	    if (argv[i][0]=='-' && argv[i][1]=='V') {
		printf("treadahead version %d.%d\n", VERSION_MAJOR, VERSION_MINOR);
		continue;
	    }
	    if (argv[i][0]=='-' && argv[i][1]=='e') {
		pid = fork();
		if (pid<0) {
			printf("Problem forking\n");
			exit(1);
		}
		if (pid>0) { /* parent */
			exit(0);
		} else {     /* child */
			if (ioidle_flag) {
				ioprio_set_idle();
			}
			if (iolow_flag) {
				ioprio_set_be_low();
			}
				
			/* get a new process group */
			setsid();
			/* close all descriptors */
			/* could use getdtablesize() here, but 20 should work */
			for (fd=20; fd>=0; --fd) {
				close(fd);
			}
			/* redirect standard I/O */
			fd = open("/dev/null", O_RDWR);
			dup(fd);
			dup(fd);
		}
		continue;
	    }
	    if (argv[i][0]=='-' && argv[i][1]=='s') {
		save_stats(STATS_OUTFILE1);
		stats_flag=1;
		continue;
	    }
	    if (argv[i][0]=='-' && argv[i][1]=='f') {
		if (strlen(&argv[i][2])) {
			strncpy(filename_buf, &argv[i][2], MAX_PATH);
			filename_buf[MAX_PATH]=0;
		} else {
			strcpy(filename_buf, READLIST_PATH);
		}
		
		printf("opening %s\n", filename_buf);
		readlist = fopen(filename_buf, "r");
		if (readlist == NULL ) {
			perror("Couldn't open readahead file");
			exit(1);
		}
		while (1) { 
			arg_count = fscanf(readlist, "%s %s\n",
				filename_buf, area_buf );
			filename = filename_buf;
			if (filename[0]=='#') {
				/* prevent reading the same '#', in case of error, */
  				/* and never ending the loop */
				filename_buf[0] = 0;
			    	continue;
			}
			/* end if not what I expected */
			if ((arg_count != 1) && (arg_count != 2)) {
				break;
			}
			if (arg_count == 1) {
				/* operate on the whole file */
				if (rewrite_blocks) {
					do_readwrite(filename, NULL);
				} else {
					do_readahead(filename, NULL);
				}
			}
			if (arg_count == 2) {
				/* operate on the area list for the file */
				if (rewrite_blocks) {
					do_readwrite(filename, area_buf);
				} else {
					do_readahead(filename, area_buf);
				}
			}
		}
		fclose(readlist);
	    } else {
		filename = argv[i];
		if (rewrite_blocks) {
			do_readwrite(filename, NULL);
		} else {
			do_readahead(filename, NULL);
		}
	    }
	}
	if (verbose) {
		printf("Total bytes read-ahead: %lu\n", total);
	}
	if (stats_flag == 1) {
		save_stats(STATS_OUTFILE2);
	}
	if (log_flag == 1) {
		write_log("treadahead end\n");
	}
	return 0;
}
